/*
 * Copyright (C) 2013 by the University of Southern California
 * $Id: mysocket.cc 19925 Tue Feb 11 14:41:03 PST 2014 Zi Hu $
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 */
#include "mysocket.h"
#include "types.h"
#include "utility.h"
#include <netinet/tcp.h>
#include <linux/socket.h>


CSocket::CSocket()
{
    _sock=-1;
    memset(&(_addr),0,sizeof(_addr));
}

CSocket::~CSocket()
{
    if(is_alive())
        ::close(_sock);
}

CSocket::CSocket(const CSocket& other) 
{
    _sock=other._sock; 
    memcpy(&_addr, &(other._addr), sizeof(sockaddr_in)); 
}

CSocket& CSocket::operator=(const CSocket& other)
{
    _sock=other._sock; 
    memcpy(&_addr, &(other._addr), sizeof(sockaddr_in)); 
    return *this;
}

bool CSocket::close()
{
    if(is_alive())
        ::close(_sock);
    return true;
}

bool CSocket::enable_fastopen()
{
    int flag = 1;
    int result = setsockopt(_sock,            /* socket affected */
                            IPPROTO_TCP,     /* set option at TCP level */
                            TCP_FASTOPEN,     /* name of option */
                            (char *) &flag,  /* the cast is historical cruft */
                            sizeof(int));    /* length of option value */
    if (result < 0)
    {
        perror("enable TCP fastopen failed\n");
        return false;
    }
    return true;
}



bool CSocket::enable_nagle()
{
    int flag = 0;
    int result = setsockopt(_sock,            /* socket affected */
                            IPPROTO_TCP,     /* set option at TCP level */
                            TCP_NODELAY,     /* name of option */
                            (char *) &flag,  /* the cast is historical cruft */
                            sizeof(int));    /* length of option value */
    if (result < 0)
    {
        perror("enable nagle failed\n");
        return false;
    }
    return true;
}

bool CSocket::disable_nagle()
{
    int flag = 1;
    int result = setsockopt(_sock,            /* socket affected */
                            IPPROTO_TCP,     /* set option at TCP level */
                            TCP_NODELAY,     /* name of option */
                            (char *) &flag,  /* the cast is historical cruft */
                            sizeof(int));    /* length of option value */
    if (result < 0)
    {
        perror("enable nagle failed\n");
        return false;
    }
    return true;
}

bool CSocket::create(int sock_type, int protocol)
{
    //create socket
    _sock=::socket(AF_INET,sock_type,protocol);
    if(_sock==-1)
    {
        perror("socket failed\t");
        return false;
    }

    //set port reuse
    int on = 1;
    if (::setsockopt(_sock, SOL_SOCKET, SO_REUSEADDR, (const char*) &on, sizeof (on))== -1)
    {
        perror("setsockopt failed\t");
        return false;
    }
    return true;
}


//bind raw socket to a specific IP address
bool CSocket::bind_rawsock(struct sockaddr_in src)
{
    if(!is_alive())
        return false;
    int bind_ret = ::bind(_sock,(struct sockaddr *) &src,sizeof(src));
    if(bind_ret==-1)
    {
        perror("bind failed\t");
        return false;
    }
    return true;
}


bool CSocket::bind(const int port)
{
    if(!is_alive())
    return false;
    sockaddr_in servaddr;
    servaddr.sin_family = AF_INET;
    servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
    servaddr.sin_port=port;
    int bind_ret = ::bind(_sock,(struct sockaddr * ) &servaddr,sizeof(servaddr));
    if(bind_ret==-1)
    {
        perror("bind failed\t");
        return false;
    }
    else
    {
        int addr_len=sizeof(_addr);
        if(getsockname(_sock, (sockaddr *) &_addr, (socklen_t *)&addr_len) == -1)
        {
            perror("getsockname failed\t");
            return false;
        }
        return true;
    }
}

bool CSocket::listen()
{
    if(!is_alive())
        return false;
    int listen_ret = ::listen(_sock, MAX_CONNECTIONS);
    if(listen_ret==-1)
    {
        perror("listen failed\t");
        return false;
    }
    return true;

}

int CSocket::send_data_TCP(const char* send_buf, const int len)
{
    if(!is_alive())
        return -1;
    //int len=strlen(send_buf);
    int send_ret= ::send(_sock,send_buf,len,0);
    if(send_ret==-1)
    {
        perror("send error:\t");
    }
    return send_ret;
}

int CSocket::send_data_TCP_fo(const char* send_buf, const int len, struct sockaddr_in& ser_addr )
{
    if(!is_alive())
        return -1;
    int addr_len=sizeof(sockaddr_in);
    int send_ret = ::sendto(_sock, send_buf , len, MSG_FASTOPEN, (const struct sockaddr*)&ser_addr, addr_len);
    if(send_ret==-1)
    {
        perror("sendto error:\t");
    }
    return send_ret;
}


int CSocket::recv_data_TCP(char* recv_buf)
{
    if(!is_alive())
        return -1;
    char buf[MAX_BUF_SIZE];
    memset(buf,0,MAX_BUF_SIZE);
    int recv_ret= ::recv(_sock,recv_buf,MAX_BUF_SIZE-1,0);
    if(recv_ret==-1)
    {
        perror("recv error:\t");
    }
    return recv_ret;
}

int CSocket::recv_data_TCP(char* recv_buf, int max_len)
{
    if(!is_alive())
        return -1;
    int recv_ret= ::recv(_sock,recv_buf,max_len,0);
    if(recv_ret==-1)
    {
        perror("recv error:\t");
    }
    return recv_ret;
}

int CSocket::send_data_UDP(const char* send_buf, const int len, struct sockaddr_in & ser_addr )
{
    if(!is_alive())
        return -1;
    //int len=strlen(send_buf);
   
    int addr_len=sizeof(sockaddr_in);
    int send_ret = ::sendto(_sock, send_buf , len, 0, (const struct sockaddr*)&ser_addr, addr_len);
    if(send_ret==-1)
    {
        perror("sendto error:\t");
    }
    return send_ret;
}


//send a tcp packet via raw socket (sendmsg())
int CSocket::send_tcp_rawsock(struct sockaddr_in src_addr, struct sockaddr_in dst_addr1, char* org_packet, int len)
{
    struct tcphdr* p= (struct tcphdr*)org_packet;
    struct iovec iov;

    memset(&iov, 0, sizeof(struct iovec));
    
    struct msghdr m = 
    {
        &dst_addr1, sizeof(struct sockaddr_in), &iov,
        1, 0, 0, 0
    };

    ssize_t bs;
    iov.iov_base = p;
    iov.iov_len = len;

    p->check = 0;
    p->check = in_cksum_tcp(src_addr.sin_addr.s_addr, dst_addr1.sin_addr.s_addr, (unsigned short *)p, len);

    if (0> (bs = sendmsg (_sock, &m, 0))) 
    {
        perror ("ERROR: sendmsg ()");
    }
    else
    {
        printf("send tcp packet, size:%d\n",(int)bs);
    }
    memset(&m, 0, sizeof(struct msghdr));
    return bs;
}



//send an ICMP packet via raw socket (sendmsg)
int CSocket::send_icmp_rawsock(struct sockaddr_in dst_addr)
{
    struct Packet {
        struct icmphdr icmp;
        struct timeval tv;
    } p;

    struct iovec iov;
    
    struct msghdr m = {
        &dst_addr, sizeof(struct sockaddr_in), &iov,
        1, 0, 0, 0
    };

    ssize_t bs;
    p.icmp.type = ICMP_ECHO;
    p.icmp.code = 0;
    p.icmp.un.echo.id = 0;
 
    iov.iov_base = &p;
    iov.iov_len = sizeof(struct icmphdr)+ sizeof(struct timeval);
 

    p.icmp.checksum = 0;
    p.icmp.un.echo.sequence = htons (0);
    gettimeofday(&p.tv, NULL);
 
    p.icmp.checksum = in_cksum((uint16_t*)&p, iov.iov_len);
 
    if (0> (bs = sendmsg (_sock, &m, 0))) 
    {
        perror ("ERROR: sendmsg ()");
    }

    memset(&p, 0, sizeof(struct Packet));
    memset(&iov, 0, sizeof(struct iovec));
    memset(&m, 0, sizeof(struct msghdr));
    return bs;
}



int CSocket::recv_data_RAW(char* recv_buf, struct sockaddr_in & ser_addr )
{
    return recv_data_UDP(recv_buf, ser_addr);
}


int CSocket::recv_data_UDP(char* recv_buf,  struct sockaddr_in& si_other)
{
    if(!is_alive())
    {
        printf("socket is closed\n");
        return -1;
    }
    char buf[MAX_BUF_SIZE];
    memset(buf,0,MAX_BUF_SIZE);
    socklen_t addr_len=sizeof(sockaddr_in);

    int recv_ret= ::recvfrom(_sock, buf ,MAX_BUF_SIZE-1,0, (struct sockaddr*)&si_other, &addr_len);
    if(recv_ret==-1)
    {
        perror("recvfrom error:\t");
    }
    else
    {
        if(recv_ret>0)
            memcpy(recv_buf,buf,MAX_BUF_SIZE);
    }
    return recv_ret;
}


bool CSocket::accept(CSocket& new_sock)
{
    if(!is_alive())
        return false;
    sockaddr_in client_addr;
    socklen_t   addr_len=sizeof(client_addr);
    int client_sock;
    client_sock= ::accept(_sock,(sockaddr*)&client_addr,&addr_len);
    if(client_sock<=0)
        return false;
    else
    {
        new_sock._sock=client_sock;
        memcpy(&(new_sock._addr),&client_addr,sizeof(client_addr));
        return true;
    }
}

bool CSocket::set_socket_nonblock(bool nonblock)
{
    int opts;
    if(!is_alive())
        return false;
    opts = fcntl ( _sock, F_GETFL );
    if ( opts < 0 )
        return false;
    if(nonblock)
        opts = ( opts | O_NONBLOCK );
    else
        opts = ( opts & ~O_NONBLOCK );
    fcntl ( _sock,F_SETFL,opts );
    return true;
}


bool CSocket::set_socket_fill_ipheader(bool on)
{
    if(setsockopt(_sock, IPPROTO_IP, IP_HDRINCL, &on, sizeof(on)) < 0)
    {
        perror("setsockopt() for IP_HDRINCL error");
        return false;
    }
    return true;
}


bool CSocket::set_socket_bind_dev(char* dev)
{
    ifreq Interface; 
    memset(&Interface, 0, sizeof(Interface)); 
    strncpy(Interface.ifr_ifrn.ifrn_name, dev, IFNAMSIZ); 
    if (setsockopt(_sock, SOL_SOCKET, SO_BINDTODEVICE, &Interface, sizeof(Interface)) < 0) 
    { 
        perror("setsockopt() for SO_BINDTODEVICE error");
        return false;
    }
    return true;
}


bool CSocket::connect(const char* serv_host,const int port)
{
    if(!is_alive())
        return false;

    struct sockaddr_in serv_addr;
    serv_addr.sin_family = AF_INET;
    serv_addr.sin_port = htons(port);
    if(serv_host==NULL)
    {
        serv_addr.sin_addr.s_addr=htonl(INADDR_ANY);
    }
    else
    {
        serv_addr.sin_addr.s_addr=inet_addr(serv_host);
    }

    if( ::connect(_sock,(sockaddr*)&serv_addr, sizeof(serv_addr))==-1)
    {
        perror("Connect failed\t");
        return false;
    }
    else
    {
        memcpy(&(_addr),&serv_addr,sizeof(serv_addr));
        return true;
    }
}
