/*
 * Copyright (C) 2013 by the University of Southern California
 * $Id: mysocket.h 19925 Tue Feb 11 14:41:03 PST 2014 Zi Hu $
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 */
#ifndef SOCKET_H
#define SOCKET_H

#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
//#include <linux/tcp.h>
#include <fcntl.h>
#include <arpa/inet.h>
#include <unistd.h>

#define MAX_CONNECTIONS 100

#ifndef _KERNEL_FASTOPEN
    #define _KERNEL_FASTOPEN
#endif

#ifndef MSG_FASTOPEN
    #define MSG_FASTOPEN    0x20000000      /* Send data in TCP SYN */
#endif

#ifndef TCP_FASTOPEN
    #define TCP_FASTOPEN    23
#endif

class CSocket{
    public:
	int _sock;
	sockaddr_in _addr;
    public:
	CSocket();
	CSocket(const CSocket& other);

	~CSocket();

	CSocket& operator=(const CSocket& other);

	bool create(int sock_type, int protocol);
	bool disable_nagle();
	bool enable_nagle();
    bool enable_fastopen();

	bool close();

	bool bind(const int port);
	bool bind_rawsock(struct sockaddr_in src);

	bool listen();
	bool accept(CSocket& new_sock);
	bool connect(const char* serv_host,const int port);

	int send_data_TCP(const char* send_buf, const int len);
	int recv_data_TCP(char* recv_buf);
    int recv_data_TCP(char* recv_buf, int max_len);
    int send_data_TCP_fo(const char* send_buf, const int len, struct sockaddr_in& ser_addr);

	int recv_data_UDP(char* recv_buf, struct sockaddr_in & si_other);
	int send_data_UDP(const char* send_buf, const int len, struct sockaddr_in & ser_addr );

	int send_data_RAW(const char* send_buf, const int len, struct sockaddr_in & ser_addr );
	int recv_data_RAW(char* recv_buf, struct sockaddr_in & ser_addr );

	int send_icmp_rawsock(struct sockaddr_in dst_addr);
	int send_tcp_rawsock(struct sockaddr_in src_addr, struct sockaddr_in dst_addr1, char* org_packet, int len);

	bool set_socket_nonblock(bool nonblock);
	bool set_socket_fill_ipheader(bool on);
	bool set_socket_bind_dev(char* dev);
	bool is_alive(){return _sock!=-1;}
};

#endif
